package org.ow2.petals.bc.filetransfer.service;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessagingException;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.stream.StreamResult;

import org.ow2.petals.bc.filetransfer.FileTransferComponent;
import org.ow2.petals.bc.filetransfer.util.FileTransferUtils;
import org.ow2.petals.bc.filetransfer.util.TransferReport;
import org.ow2.petals.commons.threadlocal.Transformers;
import org.ow2.petals.component.framework.api.Component;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.util.StringHelper;

import static org.ow2.petals.bc.filetransfer.FileTransferConstants.COPY_MODE;
import static org.ow2.petals.bc.filetransfer.FileTransferConstants.DEFAULT_FILENAME;
import static org.ow2.petals.bc.filetransfer.FileTransferConstants.ENUM_COPY_MODE_ATTACHMENTSONLY;
import static org.ow2.petals.bc.filetransfer.FileTransferConstants.ENUM_COPY_MODE_CONTENTATTACHMENTS;
import static org.ow2.petals.bc.filetransfer.FileTransferConstants.ENUM_COPY_MODE_CONTENTONLY;
import static org.ow2.petals.bc.filetransfer.FileTransferConstants.FILE_PATTERN;
import static org.ow2.petals.bc.filetransfer.FileTransferConstants.WRITE_DIRECTORY;

public class StandardWriteService extends AbstractFileService {

    private Component component = null;

    public StandardWriteService(ConfigurationExtensions extensions, Logger logger,
            Component component) {
        super(extensions, logger);
        this.component = component;
    }

    @Override
    public void process(Exchange exchange) {
        logger.info("Process received exchange: " + exchange.getExchangeId());
        if (exchange.isProviderRole()) {
            // refuse the others patterns than InOnly, ignore is it is an ack of
            // a previous refuse
            if (!exchange.isInOnlyPattern() && ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
                logger.warning("Exchange pattern '" + exchange.getPattern() + "' not supported");
                exchange.setError(new Exception(
                        "Unsupported exchange pattern, only InOnly pattern allowed"));
            } else if (exchange.getStatus() == ExchangeStatus.ACTIVE) {
                TransferReport transferReport = null;
                try {
                    String transferMode = extensions.get(COPY_MODE);
                    if (StringHelper.isNullOrEmpty(transferMode)) {
                        transferMode = ENUM_COPY_MODE_CONTENTATTACHMENTS;
                        logger.finest("using default mode : " + transferMode);
                    }
                    if (ENUM_COPY_MODE_CONTENTONLY.equalsIgnoreCase(transferMode)) {
                        this.copySource(exchange);
                    } else if (ENUM_COPY_MODE_ATTACHMENTSONLY.equalsIgnoreCase(transferMode)) {
                        transferReport = this.copyAttachments(exchange);
                    } else if (ENUM_COPY_MODE_CONTENTATTACHMENTS.equalsIgnoreCase(transferMode)) {
                        transferReport = this.copyBoth(exchange);
                    } else {
                        throw new PEtALSCDKException("Transfer mode not recognized :"
                                + transferMode);
                    }
                } catch (Exception e) {
                    logger.log(Level.SEVERE, e.getMessage(), e);
                    exchange.setError(new Exception(e.getMessage()));
                }

                if (transferReport != null && transferReport.containsFailedTransfer()) {
                    String report = transferReport.generateRawReport();
                    logger.log(Level.SEVERE, report);
                    exchange.setError(new Exception(report));
                }
            }
        } else if (!ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
            File transferedfile = ((FileTransferComponent) component).getTransferedFile(exchange);
            if (transferedfile != null) {
                logger.info("Received acknowledgment '" + exchange.getStatus().toString()
                        + "' for transfered file '" + transferedfile.getAbsolutePath()
                        + "'. Delete it");
                if (!transferedfile.delete()) {
                    logger.log(Level.WARNING, "Failed to delete file '"
                            + transferedfile.getAbsolutePath() + "'");
                }
            }
        }
    }

    /**
     * Copy the source as a File
     * 
     * @param exchange
     * @throws MessagingException
     * @throws IOException
     *             problem while writing the file
     * @throws TransformerException
     * @throws ParserConfigurationException
     */
    private void copySource(Exchange exchange) throws MessagingException, IOException,
            TransformerException, ParserConfigurationException {

        File directory = this.getTargetDirectory();
        String fileName = this.getTargetFileName(exchange, -1);

        File target = new File(directory, fileName);
        logger.info("Source transfer to file : " + target.getAbsolutePath());

        Source source = exchange.getInMessageContentAsSource();
        StreamResult streamResult = new StreamResult(new FileOutputStream(target));
        Transformer transformer = Transformers.getDefaultTransformer();
        try {
            transformer.transform(source, streamResult);
        } finally {
            transformer.reset();
            streamResult.getOutputStream().close();
        }
    }

    /**
     * Copy the attachments
     * 
     * @param exchange
     * @throws ParserConfigurationException
     * @throws FileNotFoundException
     * @throws IOException
     */
    private TransferReport copyAttachments(Exchange exchange) throws ParserConfigurationException,
            FileNotFoundException {

        File directory = this.getTargetDirectory();

        TransferReport transferReport = new TransferReport();
        Set<?> names = exchange.getInMessage().getAttachmentNames();
        long time = System.nanoTime();
        for (Object name : names) {
            DataHandler attachment = exchange.getInMessage().getAttachment(name.toString());
            this.copyAttachment(attachment, name.toString(), time, directory, transferReport);
        }

        return transferReport;
    }

    private void copyAttachment(final DataHandler attachment, final String name, final long time,
            final File directory, final TransferReport transferReport) {

        String filename = this.refactorFileName(name, time);
        File target = new File(directory, filename);
        FileOutputStream os;
        try {
            os = new FileOutputStream(target);
            logger.info("Attachment transfer to file : " + target.getAbsolutePath());
            attachment.writeTo(os);
            os.close();
            transferReport.addTransferedFile(target, null);
        } catch (IOException e) {
            logger.warning(e.getMessage());
            transferReport.addTransferedFile(target, e);
        }

    }

    /**
     * Copy the source and the attachments as files
     * 
     * @param exchange
     * 
     * @throws ParserConfigurationException
     * @throws FileNotFoundException
     * @throws MessagingException
     * @throws TransformerFactoryConfigurationError
     */
    private TransferReport copyBoth(Exchange exchange) throws ParserConfigurationException,
            FileNotFoundException, MessagingException, TransformerFactoryConfigurationError {

        TransferReport transferReport = new TransferReport();

        File directory = this.getTargetDirectory();
        long time = System.nanoTime();
        String fileName = this.getTargetFileName(exchange, time);

        // transfer the source
        File target = new File(directory, fileName);
        logger.info("Source transfer to file : " + target.getAbsolutePath());

        Source source = exchange.getInMessageContentAsSource();
        StreamResult streamResult = new StreamResult(new FileOutputStream(target));
        Transformer transformer = Transformers.getDefaultTransformer();
        try {
            transformer.transform(source, streamResult);
            streamResult.getOutputStream().close();
        } catch (TransformerException e) {
            transferReport.addTransferedFile(target, e);
            try {
                streamResult.getOutputStream().close();
            } catch (IOException ex) {
                // do nothing
            }
        } catch (IOException e) {
            transferReport.addTransferedFile(target, e);
        } finally {
            transformer.reset();
        }

        // transfer the attachments
        Set<?> names = exchange.getInMessage().getAttachmentNames();
        for (Object name : names) {
            DataHandler attachment = exchange.getInMessage().getAttachment(name.toString());
            this.copyAttachment(attachment, name.toString(), time, directory, transferReport);
        }

        return transferReport;

    }

    /**
     * Get the target directory.
     * 
     * @param workingParameter
     */
    private File getTargetDirectory() throws FileNotFoundException {
        String dir = extensions.get(WRITE_DIRECTORY);
        if (StringHelper.isNullOrEmpty(dir)) {
            dir = component.getContext().getWorkspaceRoot();
        }

        dir = FileTransferUtils.evaluateFilePath(dir);
        File directory = new File(dir);
        if (!directory.exists()) {
            throw new FileNotFoundException(WRITE_DIRECTORY + " parameter with value '" + directory
                    + "' is invalid");
        }

        return directory;
    }

    /**
     * Get the target file name for a unique transfer.
     * 
     * @return
     * @throws MessagingException
     */
    private String getTargetFileName(Exchange exchange, long time) throws MessagingException {
        String fileName = extensions.get(FILE_PATTERN);
        if (StringHelper.isNullOrEmpty(fileName)) {
            if (exchange.getOperation() != null) {
                fileName = exchange.getOperationName() + ".xml";
            } else {
                fileName = DEFAULT_FILENAME;
            }
        }
        fileName = refactorFileName(fileName, time);
        return fileName;
    }

    /**
     * Add the specified timestamp as suffix of the filename (-1 stands for "use current time")
     * 
     * @param filename
     *      The original file name
     * @param time
     *      The timestamp to append (-1 stands for "use current time")
     * @return a file name with the current date
     */
    protected String refactorFileName(String filename, long time) {
        return FileTransferUtils.addDateToFileName(filename, time);
    }
}
